@@ -2,10 +2,14 @@ package hermit
 
 import (
 	"bytes"
+	"crypto/sha256"
 	"embed"
+	"encoding/hex"
 	"fmt"
+	"io"
 	"io/fs"
 	"io/ioutil"
+	"net/http"
 	"os"
 	"path/filepath"
 	"runtime"
@@ -16,23 +20,35 @@ import (
 
 	"github.com/alecthomas/hcl"
 	"github.com/kballard/go-shellquote"
-	"github.com/pkg/errors"
 
+	"github.com/cashapp/hermit/cache"
 	"github.com/cashapp/hermit/envars"
-	"github.com/cashapp/hermit/state"
-
+	"github.com/cashapp/hermit/errors"
+	"github.com/cashapp/hermit/internal/system"
 	"github.com/cashapp/hermit/manifest"
+	"github.com/cashapp/hermit/platform"
 	"github.com/cashapp/hermit/shell"
 	"github.com/cashapp/hermit/sources"
+	"github.com/cashapp/hermit/state"
 	"github.com/cashapp/hermit/ui"
 	"github.com/cashapp/hermit/util"
 )
 
 // Reset environment variables.
 var (
+	//go:embed files/script.sha256
+	sha256Db string
+
+	// ScriptSHAs contains the default known valid SHA256 sums for bin/activate-hermit and bin/hermit.
+	ScriptSHAs = tidySha256Db(sha256Db)
+
 	//go:embed files
 	files embed.FS
 
+	// InstallerTemplateSource is a string containing the installer template
+	//go:embed "files/install.sh.tmpl"
+	InstallerTemplateSource string
+
 	// UserStateDir should be passed to Open()/Init() in most cases.
 	UserStateDir = func() string {
 		// Check if state dir is explicitly set
@@ -40,10 +56,9 @@ var (
 		if explicit != "" {
 			return explicit
 		}
-
-		cache, err := os.UserCacheDir()
+		cache, err := system.UserCacheDir()
 		if err != nil {
-			panic(err)
+			panic(fmt.Sprintf("could not find user cache dir: %s", err))
 		}
 		return filepath.Join(cache, "hermit")
 	}()
@@ -72,9 +87,10 @@ const (
 
 // Config for a Hermit environment.
 type Config struct {
-	Envars    envars.Envars `hcl:"env,optional" help:"Extra environment variables."`
-	Sources   []string      `hcl:"sources,optional" help:"Package manifest sources."`
-	ManageGit bool          `hcl:"manage-git,optional" default:"true" help:"Whether Hermit should automatically 'git add' new packages."`
+	Envars      envars.Envars `hcl:"env,optional" help:"Extra environment variables."`
+	Sources     []string      `hcl:"sources,optional" help:"Package manifest sources."`
+	ManageGit   bool          `hcl:"manage-git,optional" default:"true" help:"Whether Hermit should automatically 'git add' new packages."`
+	AddIJPlugin bool          `hcl:"idea,optional" default:"false" help:"Whether Hermit should automatically add the IntelliJ IDEA plugin."`
 }
 
 // Env is a Hermit environment.
@@ -86,18 +102,25 @@ type Env struct {
 	ephemeralEnvars envars.Ops
 	config          *Config
 	configFile      string
+	httpClient      *http.Client
+	scriptSums      []string
 
 	// Lazily initialized fields
-	lazyResolver *manifest.Resolver
-	lazySources  *sources.Sources
+	lazyResolver  *manifest.Resolver
+	lazySources   *sources.Sources
+	packageSource cache.PackageSourceSelector
 }
 
+//go:embed files/externalDependencies.xml
+var extDepData []byte
+
 // Init a new Env.
-func Init(l *ui.UI, env string, distURL string, stateDir string, config Config) error {
+func Init(l *ui.UI, env string, distURL string, stateDir string, config Config, installScriptSHA string) error {
 	env = util.RealPath(env)
 	l.Infof("Creating new Hermit environment in %s", env)
 	vars := map[string]string{
-		"HERMIT_DEFAULT_DIST_URL": distURL,
+		"HERMIT_DEFAULT_DIST_URL":      distURL,
+		"HERMIT_INSTALL_SCRIPT_SHA256": installScriptSHA,
 	}
 	bin := filepath.Join(env, "bin")
 	if err := os.Mkdir(bin, os.ModePerm); err != nil && !os.IsExist(err) {
@@ -120,6 +143,31 @@ func Init(l *ui.UI, env string, distURL string, stateDir string, config Config)
 		}
 	}
 
+	if config.AddIJPlugin {
+		ideaPath := filepath.Join(env, ".idea")
+		extDepPath := filepath.Join(ideaPath, "externalDependencies.xml")
+
+		if _, err := os.Stat(extDepPath); errors.Is(err, os.ErrNotExist) {
+			if err := os.MkdirAll(ideaPath, os.ModePerm); err != nil && !os.IsExist(err) {
+				return errors.WithStack(err)
+			}
+
+			err = ioutil.WriteFile(extDepPath, extDepData, 0600)
+			if err != nil {
+				return errors.WithStack(err)
+			}
+
+			if useGit {
+				if err = util.RunInDir(b, env, "git", "add", "-f", extDepPath); err != nil {
+					return errors.WithStack(err)
+				}
+			}
+		} else {
+			l.Infof("IntelliJ IDEA configuration already exists; skipping adding configuration.")
+		}
+
+	}
+
 	// Create a configuration file.
 	configPath := filepath.Join(bin, "hermit.hcl")
 	if _, err := os.Stat(configPath); os.IsNotExist(err) {
@@ -165,12 +213,37 @@ func EnvDirFromProxyLink(executable string) (string, error) {
 	if err != nil {
 		return "", errors.WithStack(err)
 	}
-	last := links[len(links)-1]
+	last, err := filepath.EvalSymlinks(links[len(links)-1])
+	if err != nil {
+		return "", errors.WithStack(err)
+	}
 	if filepath.Base(last) != "hermit" {
 		return "", errors.Errorf("binary is not a Hermit symlink: %s", links[0])
 	}
-	envDir := filepath.Dir(filepath.Dir(last))
-	return envDir, nil
+	last = filepath.Dir(last)
+	if filepath.Base(last) != "bin" {
+		return "", errors.Errorf("Hermit not in a bin directory: %s", links[0])
+	}
+	last = filepath.Dir(last)
+	return last, nil
+}
+
+// FindEnvDir finds the highest priority active Hermit environment
+func FindEnvDir(binary string) (envDir string, err error) {
+	// Prefer an adjacent environment
+	envDir, err = EnvDirFromProxyLink(binary)
+	if err == nil {
+		return
+	}
+
+	// Fallback to the environment from the environment variable
+	envDir = os.Getenv("HERMIT_ENV")
+	if envDir != "" {
+		return envDir, nil
+	}
+
+	// Pass up the error from EnvDirFromProxyLink
+	return
 }
 
 func getSources(l *ui.UI, envDir string, config *Config, state *state.State, defaultSources []string) (*sources.Sources, error) {
@@ -205,7 +278,16 @@ func readConfig(configFile string) (*Config, error) {
 // OpenEnv opens a Hermit environment.
 //
 // The environment may not exist, in which case this will succeed but subsequent operations will fail.
-func OpenEnv(envDir string, state *state.State, ephemeral envars.Envars) (*Env, error) {
+//
+// "scriptSums" contains all known SHA256 checksums for "bin/hermit" and "bin/activate-hermit" scripts.
+func OpenEnv(
+	envDir string,
+	state *state.State,
+	packageSource cache.PackageSourceSelector,
+	ephemeral envars.Envars,
+	httpClient *http.Client,
+	scriptSums []string,
+) (*Env, error) {
 	binDir := filepath.Join(envDir, "bin")
 	configFile := filepath.Join(binDir, "hermit.hcl")
 	config, err := readConfig(configFile)
@@ -215,8 +297,12 @@ func OpenEnv(envDir string, state *state.State, ephemeral envars.Envars) (*Env,
 
 	useGit := config.ManageGit && isEnvAGitRepo(envDir)
 	envDir = util.RealPath(envDir)
+	if len(scriptSums) == 0 {
+		scriptSums = ScriptSHAs
+	}
 
-	e := &Env{
+	return &Env{
+		packageSource:   packageSource,
 		config:          config,
 		envDir:          envDir,
 		useGit:          useGit,
@@ -224,8 +310,9 @@ func OpenEnv(envDir string, state *state.State, ephemeral envars.Envars) (*Env,
 		binDir:          binDir,
 		configFile:      configFile,
 		ephemeralEnvars: envars.Infer(ephemeral.System()),
-	}
-	return e, nil
+		httpClient:      httpClient,
+		scriptSums:      scriptSums,
+	}, nil
 }
 
 // Root directory of the environment.
@@ -233,18 +320,55 @@ func (e *Env) Root() string {
 	return e.envDir
 }
 
-// Trigger an event.
+// Verify contains valid Hermit scripts.
+func (e *Env) Verify() error {
+next:
+	for _, path := range []string{"activate-hermit", "hermit"} {
+		path = filepath.Join(e.binDir, path)
+		hasher := sha256.New()
+		r, err := os.Open(path)
+		if os.IsNotExist(err) {
+			return errors.Wrapf(err, "%s is missing, not a Hermit environment?", path)
+		} else if err != nil {
+			return errors.WithStack(err)
+		}
+		_, err = io.Copy(hasher, r)
+		_ = r.Close()
+		if err != nil {
+			return errors.WithStack(err)
+		}
+		hash := hex.EncodeToString(hasher.Sum(nil))
+		for _, candidate := range e.scriptSums {
+			if hash == candidate {
+				continue next
+			}
+		}
+		return errors.Errorf("%s has an unknown SHA256 signature (%s); verify that you trust this environment and run 'hermit init %s'", path, hash, e.envDir)
+	}
+	return nil
+}
+
+// Trigger an event for all installed packages.
 func (e *Env) Trigger(l *ui.UI, event manifest.Event) (messages []string, err error) {
 	pkgs, err := e.ListInstalled(l)
 	if err != nil {
 		return nil, err
 	}
 	for _, pkg := range pkgs {
-		eventMessages, err := pkg.Trigger(l, event)
+		pkgMessages, err := e.TriggerForPackage(l, event, pkg)
 		if err != nil {
-			return nil, errors.Wrapf(err, "%s: on %s", pkg, event)
+			return nil, err
 		}
-		messages = append(messages, eventMessages...)
+		messages = append(messages, pkgMessages...)
+	}
+	return messages, nil
+}
+
+// TriggerForPackage triggers an event for a single package.
+func (e *Env) TriggerForPackage(l *ui.UI, event manifest.Event, pkg *manifest.Package) (messages []string, err error) {
+	messages, err = pkg.Trigger(l, event)
+	if err != nil {
+		return nil, errors.Wrapf(err, "%s: on %s", pkg, event)
 	}
 	return messages, nil
 }
@@ -261,11 +385,6 @@ func (e *Env) ValidateManifests(l *ui.UI) (manifest.ManifestErrors, error) {
 	return resolver.Errors(), nil
 }
 
-// GC can be used to clean up unused packages, and clear the download cache.
-func (e *Env) GC(l *ui.UI, age time.Duration) error {
-	return e.state.GC(l, age, e.Resolve)
-}
-
 // LinkedBinaries lists just the binaries installed in the environment.
 func (e *Env) LinkedBinaries(pkg *manifest.Package) (binaries []string, err error) {
 	files, err := ioutil.ReadDir(e.binDir)
@@ -311,11 +430,6 @@ func (e *Env) uninstall(l *ui.Task, pkg *manifest.Package) (*shell.Changes, erro
 	changes := shell.NewChanges(envars.Parse(os.Environ()))
 	changes.Remove = ops
 
-	err = e.state.RecordUninstall(pkg, e.binDir)
-	if err != nil {
-		return nil, errors.WithStack(err)
-	}
-
 	return changes, nil
 }
 
@@ -388,9 +502,17 @@ func (e *Env) Test(l *ui.UI, pkg *manifest.Package) error {
 	if !found {
 		return errors.Errorf("couldn't find test executable %q in package %s", args[0], pkg)
 	}
-	cmd, _ := util.Command(task, args...)
-	cmd.Env = e.allEnvarsForPackages(true, pkg)
-	return cmd.Run()
+	cmd, out := util.Command(task, args...)
+	deps, err := e.ensureRuntimeDepsPresent(l, pkg)
+	if err != nil {
+		return errors.WithStack(err)
+	}
+	cmd.Env = e.envarsFromOps(true, e.allEnvarOpsForPackages(deps, pkg))
+	err = cmd.Run()
+	if err != nil {
+		return errors.Wrap(err, out.String())
+	}
+	return nil
 }
 
 // Unpack but do not install package.
@@ -407,9 +529,13 @@ func (e *Env) Install(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
 	if err != nil {
 		return nil, errors.WithStack(err)
 	}
+	if err := pkg.EnsureSupported(); err != nil {
+		return nil, errors.Wrapf(err, "install failed")
+	}
 
 	allChanges := shell.NewChanges(envars.Parse(os.Environ()))
 
+	didUninstall := false
 	for _, ipkg := range installed {
 		if ipkg.Reference.Name == pkg.Reference.Name {
 			changes, err := e.uninstall(task, ipkg)
@@ -417,10 +543,22 @@ func (e *Env) Install(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
 				return nil, errors.WithStack(err)
 			}
 			allChanges = allChanges.Merge(changes)
+			didUninstall = true
 		}
 	}
 
-	changes, err := e.install(task, pkg)
+	if !didUninstall && len(pkg.UnsupportedPlatforms) > 0 {
+
+		resp, err := l.Confirmation("%s is not supported on these Hermit platforms: %s, are you sure you want to install it? [y/N]", pkg.Reference, pkg.UnsupportedPlatforms)
+		if err != nil {
+			return nil, errors.WithStack(err)
+		}
+		if !resp {
+			return allChanges, nil
+		}
+	}
+
+	changes, err := e.install(l, pkg)
 	if err != nil {
 		return nil, errors.WithStack(err)
 	}
@@ -428,39 +566,109 @@ func (e *Env) Install(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
 	return allChanges.Merge(changes), nil
 }
 
+// resolveRuntimeDependencies checks all runtime dependencies for a package are available.
+//
+// Aggregate and collect the package names and binaries of all runtime dependencies to avoid collisions.
+func (e *Env) resolveRuntimeDependencies(l *ui.UI, p *manifest.Package, aggregate map[string]*manifest.Package, bins map[string]*manifest.Package) error {
+	var depPkgs []*manifest.Package
+	// Explicitly specified runtime-dependencies in the package.
+	for _, ref := range p.RuntimeDeps {
+		previous := aggregate[ref.Name]
+		if previous != nil && previous.Reference.Compare(ref) != 0 {
+			return errors.Errorf("two conflicting runtime-dependencies: %s vs %s", ref, previous.Reference)
+		}
+
+		depPkg, err := e.Resolve(l, manifest.ExactSelector(ref), true)
+		if err != nil {
+			return errors.WithStack(err)
+		}
+
+		depPkgs = append(depPkgs, depPkg)
+	}
+
+	for _, depPkg := range depPkgs {
+		for _, bin := range depPkg.Binaries {
+			base := filepath.Base(bin)
+			if previous, ok := bins[base]; ok && !previous.Reference.Match(depPkg.Reference) {
+				return errors.Errorf("conflicting binary %q in multiple runtime dependencies: %s and %s", bin, previous.Reference, depPkg.Reference)
+			}
+			bins[base] = depPkg
+		}
+
+		aggregate[depPkg.Reference.Name] = depPkg
+		if err := e.resolveRuntimeDependencies(l, depPkg, aggregate, bins); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
+func (e *Env) ensureRuntimeDepsPresent(l *ui.UI, p *manifest.Package) ([]*manifest.Package, error) {
+	deps := map[string]*manifest.Package{}
+	err := e.resolveRuntimeDependencies(l, p, deps, map[string]*manifest.Package{})
+	if err != nil {
+		return nil, errors.WithStack(err)
+	}
+	result := make([]*manifest.Package, 0, len(deps))
+	for _, pkg := range deps {
+		if err := e.state.CacheAndUnpack(l.Task(p.Reference.String()), pkg); err != nil {
+			return nil, errors.WithStack(err)
+		}
+		result = append(result, pkg)
+	}
+	return result, nil
+}
+
+// Update timestamps for runtime dependencies.
+func (e *Env) writePackageState(pkgs ...*manifest.Package) error {
+	for _, pkg := range pkgs {
+		if err := e.state.WritePackageState(pkg); err != nil {
+			return errors.WithStack(err)
+		}
+	}
+	return nil
+}
+
 // install a package
-func (e *Env) install(l *ui.Task, p *manifest.Package) (*shell.Changes, error) {
+func (e *Env) install(l *ui.UI, p *manifest.Package) (*shell.Changes, error) {
+	task := l.Task(p.Reference.String())
+
+	pkgs, err := e.ensureRuntimeDepsPresent(l, p)
+	if err != nil {
+		return nil, errors.WithStack(err)
+	}
 	p.UpdatedAt = time.Now()
-	log := l.SubTask("install")
+	log := task.SubTask("install")
 	log.Infof("Installing %s", p)
 	log.Debugf("From %s", p.Source)
 	log.Debugf("To %s", p.Dest)
-	err := e.state.CacheAndUnpack(l, p)
+	err = e.state.CacheAndUnpack(task, p)
 	if err != nil {
 		return nil, errors.WithStack(err)
 	}
 	if _, err := os.Stat(e.pkgLink(p)); os.IsNotExist(err) {
-		if err = e.linkPackage(l, p); err != nil {
-			return nil, errors.WithStack(err)
-		}
-		if err = e.state.WritePackageState(p, e.binDir); err != nil {
+		if err = e.linkPackage(task, p); err != nil {
 			return nil, errors.WithStack(err)
 		}
+		pkgs = append(pkgs, p)
 	}
 	ops := e.envarsForPackages(p)
 	changes := shell.NewChanges(envars.Parse(os.Environ()))
 	changes.Add = ops
 
-	return changes, nil
+	return changes, errors.WithStack(e.writePackageState(pkgs...))
 }
 
 // Upgrade package.
-func (e *Env) Upgrade(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
+//
+// If an upgrade does not occur, returns all nils.
+func (e *Env) Upgrade(l *ui.UI, pkg *manifest.Package) (*shell.Changes, *manifest.Package, error) {
 	task := l.Task(pkg.Reference.String())
 
 	if pkg.Reference.IsChannel() {
 		err := e.state.UpgradeChannel(task, pkg)
-		return nil, errors.WithStack(err)
+		return nil, nil, errors.WithStack(err)
 	}
 	return e.upgradeVersion(l, pkg)
 }
@@ -469,7 +677,7 @@ func (e *Env) Upgrade(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
 //
 // Link chains are in the form
 //
-//     <binary> -> <pkg>-<version>.pkg -> hermit
+//	<binary> -> <pkg>-<version>.pkg -> hermit
 func (e *Env) ResolveLink(l *ui.UI, executable string) (pkg *manifest.Package, binary string, err error) {
 	links, err := util.ResolveSymlinks(executable)
 	if err != nil {
@@ -523,7 +731,10 @@ func (e *Env) Exec(l *ui.UI, pkg *manifest.Package, binary string, args []string
 	if err != nil {
 		return errors.WithStack(err)
 	}
-	err = e.EnsureChannelIsUpToDate(l, pkg)
+	if err := e.EnsureChannelIsUpToDate(l, pkg); err != nil {
+		return errors.WithStack(err)
+	}
+	runtimeDeps, err := e.ensureRuntimeDepsPresent(l, pkg)
 	if err != nil {
 		return errors.WithStack(err)
 	}
@@ -531,7 +742,21 @@ func (e *Env) Exec(l *ui.UI, pkg *manifest.Package, binary string, args []string
 	if err != nil {
 		return errors.WithStack(err)
 	}
-	env, err := e.Envars(l, true)
+
+	installed, err := e.ListInstalled(l)
+	if err != nil {
+		return errors.WithStack(err)
+	}
+	ops := e.allEnvarOpsForPackages(runtimeDeps, installed...)
+	packageHermitBin, err := e.getPackageRuntimeEnvops(pkg)
+	if err != nil {
+		return errors.WithStack(err)
+	}
+	if packageHermitBin != nil {
+		ops = append(ops, packageHermitBin)
+	}
+	env := e.envarsFromOps(true, ops)
+
 	if err != nil {
 		return errors.WithStack(err)
 	}
@@ -539,20 +764,34 @@ func (e *Env) Exec(l *ui.UI, pkg *manifest.Package, binary string, args []string
 		if filepath.Base(bin) != filepath.Base(binary) {
 			continue
 		}
-		err = e.state.WritePackageState(pkg, e.binDir)
-		if err != nil {
-			return errors.WithStack(err)
-		}
-		b.Tracef("exec %s", shellquote.Join(append([]string{bin}, args...)...))
+		argsCopy := make([]string, len(args))
+		copy(argsCopy, args)
+		argsCopy[0] = bin
+		b.Tracef("exec %s", shellquote.Join(argsCopy...))
 		l.Clear()
 		timer()
 
-		err = syscall.Exec(bin, args, env)
+		err = syscall.Exec(bin, argsCopy, env)
 		return errors.Wrapf(err, "%s: failed to execute %q", pkg, bin)
 	}
 	return errors.Errorf("%s: could not find binary %q", pkg, binary)
 }
 
+func (e *Env) getPackageRuntimeEnvops(pkg *manifest.Package) (envars.Op, error) {
+	// If the package contains a Hermit env, add that to the PATH for runtime dependencies
+	pkgEnv, err := OpenEnv(pkg.Root, e.state, e.packageSource, nil, e.httpClient, e.scriptSums)
+	if err != nil {
+		return nil, errors.WithStack(err)
+	}
+	if err = pkgEnv.Verify(); err != nil {
+		if !errors.Is(err, os.ErrNotExist) {
+			return nil, errors.WithStack(err)
+		}
+		return nil, nil
+	}
+	return &envars.Prepend{Name: "PATH", Value: pkgEnv.binDir}, nil
+}
+
 // Resolve package reference.
 //
 // If "syncOnMissing" is true, sources will be synced if the selector cannot
@@ -578,6 +817,80 @@ func (e *Env) Resolve(l *ui.UI, selector manifest.Selector, syncOnMissing bool)
 	return resolved, nil
 }
 
+// ValidationOptions for manifest validation
+type ValidationOptions struct {
+	// CheckSources if true, check that the package sources are reachable
+	CheckSources bool
+}
+
+// ValidateManifest with given name.
+//
+// Returns the resolution errors for core systems as warnings.
+// If a version fails to resolve for all systems, returns an error.
+func (e *Env) ValidateManifest(l *ui.UI, name string, options *ValidationOptions) ([]string, error) {
+	sources, err := e.sources(l)
+	if err != nil {
+		return nil, errors.WithStack(err)
+	}
+
+	mnf, err := manifest.NewLoader(sources).Load(l, name)
+	if err != nil {
+		return nil, errors.WithStack(err)
+	}
+
+	refs := mnf.References(name)
+	var warnings []string
+	task := l.Task("validate")
+	for _, ref := range refs {
+		task.Infof("Validating %s", ref)
+		w, err := e.validateReference(l, sources, ref, options)
+		if err != nil {
+			return nil, errors.WithStack(err)
+		}
+		warnings = append(warnings, w...)
+	}
+
+	return warnings, nil
+}
+
+func (e *Env) validateReference(l *ui.UI, srcs *sources.Sources, ref manifest.Reference, options *ValidationOptions) ([]string, error) {
+	var fails []string
+	var warnings []string
+	for _, p := range platform.Core {
+		resolver, err := manifest.New(srcs, manifest.Config{
+			Env:   e.envDir,
+			State: e.state.Root(),
+			Platform: platform.Platform{
+				OS:   p.OS,
+				Arch: p.Arch,
+			},
+		})
+		if err != nil {
+			return nil, errors.WithStack(err)
+		}
+
+		pkg, err := resolver.Resolve(l, manifest.ExactSelector(ref))
+		if err != nil {
+			msg := fmt.Sprintf("%s: %s", p, err.Error())
+			fails = append(fails, msg)
+			warnings = append(warnings, msg)
+			continue
+		}
+
+		if options.CheckSources {
+			if err := manifest.ValidatePackageSource(e.packageSource, e.httpClient, pkg.Source); err != nil {
+				return nil, errors.Wrapf(err, "%s: %s", ref, p)
+			}
+		}
+
+		warnings = append(warnings, pkg.Warnings...)
+	}
+	if len(fails) >= len(platform.Core) {
+		return nil, errors.Errorf("%s failed to resolve on all platforms: %s", ref, strings.Join(fails, "; "))
+	}
+	return warnings, nil
+}
+
 // ResolveVirtual references to concrete packages.
 func (e *Env) ResolveVirtual(l *ui.UI, name string) ([]*manifest.Package, error) {
 	resolver, err := e.resolver(l)
@@ -597,7 +910,7 @@ func (e *Env) ResolveVirtual(l *ui.UI, name string) ([]*manifest.Package, error)
 // UpdateUsage updates the package usage time stamps in the underlying database.
 // if the package was not previously present, it is inserted to the DB.
 func (e *Env) UpdateUsage(pkg *manifest.Package) error {
-	err := e.state.WritePackageState(pkg, e.binDir)
+	err := e.state.WritePackageState(pkg)
 	if err != nil {
 		return errors.WithStack(err)
 	}
@@ -605,6 +918,8 @@ func (e *Env) UpdateUsage(pkg *manifest.Package) error {
 }
 
 // ListInstalledReferences from this environment.
+//
+// This function is much faster than ListInstalled, if all you need is the Reference.
 func (e *Env) ListInstalledReferences() ([]manifest.Reference, error) {
 	matches, err := filepath.Glob(filepath.Join(e.binDir, ".*.pkg"))
 	if err != nil {
@@ -646,11 +961,11 @@ func (e *Env) ListInstalled(l *ui.UI) ([]*manifest.Package, error) {
 // only variables defined in Hermit itself will be available.
 func (e *Env) Envars(l *ui.UI, inherit bool) ([]string, error) {
 	defer ui.LogElapsed(l, "envars")()
-	pkgs, err := e.ListInstalled(l)
+	ops, err := e.EnvOps(l)
 	if err != nil {
 		return nil, err
 	}
-	return e.allEnvarsForPackages(inherit, pkgs...), nil
+	return e.envarsFromOps(inherit, ops), nil
 }
 
 // EnvOps returns the envar mutation operations for this environment.
@@ -663,11 +978,7 @@ func (e *Env) EnvOps(l *ui.UI) (envars.Ops, error) {
 	if err != nil {
 		return nil, err
 	}
-	ops := e.envarsForPackages(pkgs...)
-	ops = append(ops, e.hermitEnvarOps()...)
-	ops = append(ops, e.localEnvarOps()...)
-	ops = append(ops, e.ephemeralEnvars...)
-	return ops, nil
+	return e.allEnvarOpsForPackages(nil, pkgs...), nil
 }
 
 // SetEnv sets an extra environment variable.
@@ -765,7 +1076,7 @@ func (e *Env) AddSource(l *ui.UI, s sources.Source) error {
 		return errors.WithStack(err)
 	}
 	sources.Add(s)
-	return e.Sync(l, true)
+	return e.Update(l, true)
 }
 
 // EnvDir returns the directory where this environment is rooted
@@ -781,31 +1092,33 @@ func (e *Env) BinDir() string {
 // upgradeVersion upgrades the package to its latest version.
 //
 // If the package is already at its latest version, this is a no-op.
-func (e *Env) upgradeVersion(l *ui.UI, pkg *manifest.Package) (*shell.Changes, error) {
+func (e *Env) upgradeVersion(l *ui.UI, pkg *manifest.Package) (*shell.Changes, *manifest.Package, error) {
 	resolver, err := e.resolver(l)
 	if err != nil {
-		return nil, errors.WithStack(err)
+		return nil, nil, errors.WithStack(err)
 	}
 	// Get the latest version of the package
 	resolved, err := resolver.Resolve(l, manifest.PrefixSelector(pkg.Reference.Major()))
 	if err != nil {
-		return nil, errors.WithStack(err)
+		return nil, nil, errors.WithStack(err)
+	}
+	if err := resolved.EnsureSupported(); err != nil {
+		return nil, nil, errors.Wrapf(err, "upgrade failed")
 	}
 	if !resolved.Reference.Version.Match(pkg.Reference.Version) {
 		l.Task(pkg.Reference.Name).SubTask("upgrade").Infof("Upgrading %s to %s", pkg, resolved)
 		uc, err := e.uninstall(l.Task(pkg.Reference.String()), pkg)
 		if err != nil {
-			return nil, errors.WithStack(err)
+			return nil, nil, errors.WithStack(err)
 		}
-		ic, err := e.install(l.Task(resolved.Reference.String()), resolved)
+		ic, err := e.install(l, resolved)
 		if err != nil {
-			return nil, errors.WithStack(err)
+			return nil, nil, errors.WithStack(err)
 		}
 		// Update the package.
-		*pkg = *resolved
-		return uc.Merge(ic), nil
+		return uc.Merge(ic), resolved, nil
 	}
-	return shell.NewChanges(envars.Parse(os.Environ())), nil
+	return nil, nil, nil
 }
 
 func (e *Env) readPackageState(pkg *manifest.Package) {
@@ -890,15 +1203,20 @@ func (e *Env) pkgLink(pkg *manifest.Package) string {
 }
 
 // Returns combined system + Hermit + package environment variables, fully expanded.
-//
-// If "inherit" is true, system envars will be included.
-func (e *Env) allEnvarsForPackages(inherit bool, pkgs ...*manifest.Package) []string {
+func (e *Env) allEnvarOpsForPackages(runtimeDeps []*manifest.Package, pkgs ...*manifest.Package) envars.Ops {
 	var ops envars.Ops
-	system := envars.Parse(os.Environ())
+	ops = append(ops, e.hermitEnvarOps()...)
+	ops = append(ops, e.hermitRuntimeDepOps(runtimeDeps)...)
 	ops = append(ops, e.envarsForPackages(pkgs...)...)
 	ops = append(ops, e.localEnvarOps()...)
-	ops = append(ops, e.hermitEnvarOps()...)
 	ops = append(ops, e.ephemeralEnvars...)
+	return ops
+}
+
+// Converts given envars.Ops into actual environment variables
+// If "inherit" is true, system envars will be included.
+func (e *Env) envarsFromOps(inherit bool, ops envars.Ops) []string {
+	system := envars.Parse(os.Environ())
 	transform := system.Apply(e.Root(), ops)
 	if inherit {
 		return transform.Combined().System()
@@ -920,7 +1238,7 @@ func (e *Env) localEnvarOps() envars.Ops {
 	return envars.Infer(e.config.Envars.System())
 }
 
-// hermitEnvarOps returns the environment variables created and reuqired by hermit itself
+// hermitEnvarOps returns the environment variables created and required by hermit itself
 func (e *Env) hermitEnvarOps() envars.Ops {
 	return envars.Ops{
 		&envars.Prepend{Name: "PATH", Value: e.binDir},
@@ -929,6 +1247,15 @@ func (e *Env) hermitEnvarOps() envars.Ops {
 	}
 }
 
+// hermitRuntimeDepOps returns the environment variables for runtime dependencies
+func (e *Env) hermitRuntimeDepOps(pkgs []*manifest.Package) envars.Ops {
+	ops := e.envarsForPackages(pkgs...)
+	for _, pkg := range pkgs {
+		ops = append(ops, &envars.Prepend{Name: "PATH", Value: filepath.Join(e.state.BinaryDir(), pkg.Reference.String())})
+	}
+	return ops
+}
+
 func (e *Env) linkApp(app string) error {
 	root := filepath.Join(e.binDir, filepath.Base(app))
 	for _, dir := range []string{"Contents/MacOS", "Contents/Resources"} {
@@ -940,13 +1267,36 @@ func (e *Env) linkApp(app string) error {
 	return nil
 }
 
-// Sync sources.
-func (e *Env) Sync(l *ui.UI, force bool) error {
+// Update sources and auto-update channels.
+//
+// Will be updated at most every SyncFrequency unless "force" is true.
+//
+// A Sources set can only be synchronised once. Following calls will not have any effect.
+func (e *Env) Update(l *ui.UI, force bool) error {
 	resolver, err := e.resolver(l)
 	if err != nil {
 		return errors.WithStack(err)
 	}
-	return errors.WithStack(resolver.Sync(l, force))
+	if err := resolver.Sync(l, force); err != nil {
+		return errors.WithStack(err)
+	}
+	pkgs, err := e.ListInstalled(l)
+	if err != nil {
+		return errors.WithStack(err)
+	}
+	for _, pkg := range pkgs {
+		if pkg.Reference.IsChannel() {
+			log := l.Task(pkg.String())
+			if force || time.Since(pkg.UpdatedAt) > pkg.UpdateInterval {
+				if err := e.state.UpgradeChannel(log, pkg); err != nil {
+					return errors.Wrap(err, pkg.String())
+				}
+			} else {
+				log.Debugf("Update skipped, updated within the last %s", pkg.UpdateInterval)
+			}
+		}
+	}
+	return nil
 }
 
 // Sources enabled in this environment.
@@ -959,13 +1309,13 @@ func (e *Env) Sources(l *ui.UI) ([]string, error) {
 }
 
 // ResolveWithDeps collect packages and their dependencies based on the given manifest.Selector into a map
-func (e *Env) ResolveWithDeps(l *ui.UI, installed manifest.Packages, selector manifest.Selector, out map[string]*manifest.Package) (err error) {
+func (e *Env) ResolveWithDeps(l *ui.UI, installed []manifest.Reference, selector manifest.Selector, out map[string]*manifest.Package) (err error) {
 	for _, existing := range installed {
-		if existing.Reference.String() == selector.Name() {
+		if existing.String() == selector.Name() {
 			return nil
 		}
 	}
-	pkg, err := e.Resolve(l, selector, false)
+	pkg, err := e.Resolve(l, selector, true)
 	if err != nil {
 		return errors.WithStack(err)
 	}
@@ -975,7 +1325,11 @@ func (e *Env) ResolveWithDeps(l *ui.UI, installed manifest.Packages, selector ma
 		ref, err := e.resolveVirtual(l, req)
 		if err != nil && errors.Is(err, manifest.ErrUnknownPackage) {
 			// Secondly search by the package name
-			if err = e.ResolveWithDeps(l, installed, manifest.NameSelector(req), out); err != nil {
+			sel, err := manifest.ParseGlobSelector(req)
+			if err != nil {
+				return errors.WithStack(err)
+			}
+			if err = e.ResolveWithDeps(l, installed, sel, out); err != nil {
 				return errors.WithStack(err)
 			}
 		} else if err != nil {
@@ -1008,6 +1362,7 @@ func (e *Env) resolveVirtual(l *ui.UI, name string) (manifest.Reference, error)
 			}
 		}
 	}
+	sort.Strings(candidates)
 	return manifest.Reference{}, errors.Errorf("multiple packages satisfy the required dependency %q, please install one of the following manually: %s", name, strings.Join(candidates, ", "))
 }
 
@@ -1060,8 +1415,10 @@ func (e *Env) resolver(l *ui.UI) (*manifest.Resolver, error) {
 	resolver, err := manifest.New(sources, manifest.Config{
 		Env:   e.envDir,
 		State: e.state.Root(),
-		OS:    runtime.GOOS,
-		Arch:  runtime.GOARCH,
+		Platform: platform.Platform{
+			OS:   runtime.GOOS,
+			Arch: runtime.GOARCH,
+		},
 	})
 	if err != nil {
 		return nil, errors.WithStack(err)
@@ -1069,3 +1426,23 @@ func (e *Env) resolver(l *ui.UI) (*manifest.Resolver, error) {
 	e.lazyResolver = resolver
 	return resolver, nil
 }
+
+// tidySha256Db is a helper function to remove leading and trailing
+// whitespaces and empty lines from a multiline string containing
+// SHA-256 hash digests, and return a slice of digest strings that do
+// not start with a '#' character (used as a comment line indicator).
+func tidySha256Db(s string) []string {
+	ss := strings.Split(s, "\n")
+
+	var res []string
+	for _, x := range ss {
+		line := strings.TrimSpace(x)
+		// Add to result only when hash digest length is correct. Skip
+		// comment line.
+		if len(line) == 64 && !strings.HasPrefix(line, "#") {
+			res = append(res, line)
+		}
+	}
+
+	return res
+}
